/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  CDA 
*
* @FILE     apachecda.c

*
* @ABSTRACT
* Apache cda handler
*
* This file contains:
*
* @INCLUDE


#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"
#include "http_core.h"
#include "http_protocol.h"

#include "gcn_upcon.h"
#include "gcn_uptyp.h"
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "evr_upmac.h"

#include "com_upcon.h"
#include "com_uptyp.h"
#include "com_upext.h"

#include "apachecda_prcon.h"
#include "apachecda_upcon.h"
#include "apachecda_upext.h"

module MODULE_VAR_EXPORT cdaModule;


static CSILONG XMLEscape(request_rec *r, char *szIn, char **pszOut)
{
    size_t i;
    CSILONG lLength;
    char *szPtr;

    /* no string, no conversion */
    if (NULL == szIn)
    {
        *pszOut = NULL;
        return CSISUCCESS;
    }
     
    /* First the length of the new string is calculated */
    for (i = 0, szPtr = szIn, lLength = 1; i < strlen(szIn); i++, szPtr++)
    {
        switch(szPtr[0])
        {
            case '<' :
            case '>' :
                lLength += 4;
                break;       
            case '&' :
                lLength += 5;
                break;       
            case '\'' :
            case '\"' :
                lLength += 6;
                break;       
            default : 
                lLength++;
                break;    
        }
    }
    
    *pszOut = ap_pcalloc(r->pool, lLength); 

    /* The escaped string is calculated */
    for (i = 0, szPtr = *pszOut; i < strlen(szIn); i++)   
    {
        switch(szIn[i])
        {
            case '<' :
                strcat(szPtr, "&lt;");
                szPtr +=4;
                break;    
            case '>' :
                strcat(szPtr, "&gt;");
                szPtr +=4;
                break;    
            case '&' :
                strcat(szPtr, "&amp;");
                szPtr +=5;
                break;    
            case '\'' :
                strcat(szPtr, "&apos;");
                szPtr +=6;
                break;    
            case '\"' :
                strcat(szPtr, "&quot;");
                szPtr +=6;
                break;    
            default : 
                szPtr[0] = szIn[i];
                szPtr++;
                break;  
        }
    }
    return CSISUCCESS;
}


static void ErrorOutput(request_rec *r
                       ,char *szAction
                       ,long lId
                       )
{
    ap_rprintf(r
              ,XML_CDA_BEGINTAG
              ,NULL == szAction ? "none" : szAction
              ,COM_FAILURE_STR
              );              
    ap_rprintf(r
              ,XML_ERROR_BEGINTAG
              ,lId
              );                  
    ap_rprintf(r
              ,"%s=\"%s\"\n"
              ,COM_DESCRIPTION_STR
              ,APACHE_CDA_MSGS[lId]
              );             
    ap_rputs(XML_ERROR_ENDTAG, r);
    ap_rputs(XML_CDA_ENDTAG, r); 
}


static void ErrorOutputStr(request_rec *r
                          ,char *szAction
                          ,long lId
                          ,char *szErrMsg
                          )
{
    ap_rprintf(r
              ,XML_CDA_BEGINTAG
              ,NULL == szAction ? "none" : szAction
              ,COM_FAILURE_STR
              );              
    ap_rprintf(r
              ,XML_ERROR_BEGINTAG
              ,lId
              );                  
    ap_rprintf(r
              ,"%s=\"%s\"\n"
              ,COM_DESCRIPTION_STR
              ,szErrMsg
              );             
    ap_rputs(XML_ERROR_ENDTAG, r);
    ap_rputs(XML_CDA_ENDTAG, r); 
}


static int cda_manProgramChild(void *argv, child_info *info)
{
    ap_cleanup_for_exec();
    execvp(((char **) argv)[0], (char **) argv);
    return 1;
}


static int cdaHandler(request_rec *r) 
{
    table *arguments = NULL;
    const CSICHAR *szAction = NULL;
    char *argv[4];
    FILE *fpout;
    int rc;

    r->content_type = "text/html";
    ap_send_http_header(r);
    ap_rputs(XML_HEADER, r);

    /* Retrieving the arguments */
    arguments = ArgumentsParse(r);

    /* Retrieving the cda action*/
    szAction = ap_table_get(arguments, APACHE_CDA_ACTION);

    if (NULL == szAction)
    {
        ErrorOutput(r
                   ,NULL
                   ,APACHE_CDA_NO_ACTION
                   );
        return OK;
    }
    
    argv[0] = ap_pstrdup(r->pool, APACHE_CDA_MAN);
    
    if (!(strcmp(szAction, APACHE_CDA_FLUSH)))
    {
        argv[1] = ap_pstrdup(r->pool, APACHE_CDA_MAN_FLUSH);
        argv[2] = (char * ) 0;
    }
    else if (!(strcmp(szAction, APACHE_CDA_UNCACHE)))
    {
        const CSICHAR *szDN = NULL;

        szDN = ap_table_get(arguments, APACHE_CDA_DN);
        if (NULL == szDN)
        {
            ErrorOutput(r
                       ,APACHE_CDA_UNCACHE
                       ,APACHE_CDA_NO_DN
                       );
            return OK;
        }
        argv[1] = ap_pstrdup(r->pool, APACHE_CDA_MAN_UNCACHE);
        argv[2] = ap_pstrdup(r->pool, szDN);
        argv[3] = (char * ) 0;
    }
    else if (!(strcmp(szAction, APACHE_CDA_STATUS)))
    {
        argv[1] = ap_pstrdup(r->pool, APACHE_CDA_MAN_STATUS);
        argv[2] = (char * ) 0;
    }
    else
    {
        ErrorOutput(r
                   ,(char *) szAction
                   ,APACHE_CDA_BAD_ACTION
                   );
        return OK;
    }

    rc = ap_spawn_child(r->pool
                       ,cda_manProgramChild
                       ,(void *) argv
                       ,kill_after_timeout
                       ,NULL
                       ,&fpout
                       ,NULL
                       );
    if (!rc || NULL == fpout)
    {
        ErrorOutput(r
                   ,(char *) szAction
                   ,APACHE_CDA_SPAWN_ERROR
                   );
        return OK;
    }
    {
        int fout = fileno(fpout);
        char buf[2048];
        CSICHAR *escbuf;
        int i = 0;
        char c;

        while (read(fout, &c, 1) == 1 && (i < 2048 - 1)) 
        {
            if (c == '\n')
                break;
            buf[i++] = c;
        }   
        buf[i] = '\0';
        XMLEscape(r, buf, &escbuf);
/* OR7641: detect cda_man error messages for action "flush"
*          "Unable to get semaphore for write operation at index %lu.\n", i"
*          "Unable to unget semaphore for write operation at index %lu.\n", i"
*          "Failed to bind with CDA cache: errno = %d.\n", errno */
        if ((!strcmp(szAction, "flush") && i > 0))
        {
           ErrorOutputStr(r
                         ,(char *) szAction
                         ,APACHE_CDA_MAN_ERROR
                         ,(char *) escbuf
                         );
        }
        else
        {
           ap_rprintf(r
                     ,XML_CDA_BEGINTAG
                     ,(char *) szAction
                     ,COM_SUCCESS_STR
                     );
           ap_rprintf(r
                     ,"%s=\"%s\"\n"
                     ,COM_DESCRIPTION_STR
                     ,escbuf
                     );
           ap_rputs(XML_CDA_ENDTAG, r);
        }
    }

    return OK;
}
