/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  COMMON
*
* @FILE     cookie.c
*
* @ABSTRACT
* Apache common functions
*
* This file contains:
* The cookie management used by the Apache handlers
*
* @INCLUDE
*/

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"

#include "gcn_uptyp.h"
#include "gcn_upcon.h"
#include "cdx_upcon.h"
#include "cdx_uptyp.h"

#include "com_upext.h"
#include "com_uptyp.h"

array_header *CookieParse(request_rec *r)
{
    array_header *cookies = NULL;
    const CSICHAR *data = ap_table_get(r->headers_in
                                      ,"Cookie"
                                      );
    const char *pair;

/*    ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "cookie %s", data);*/

    cookies = ap_make_array(r->pool
                           ,0 /* Apache resizes it when needed */
                           ,sizeof(COOKIE *)
                           );
    while (   NULL != data
	   && *data
           && (pair = ap_getword(r->pool, &data, ';'))
          )
    {
        COOKIE **cookie;
        const char *name, *key, *value;

        /* Skipping spaces between semicolon and following cookie */
        while (ap_isspace(*data)) ++data;

        /* Pushing the cookie into the array */
        cookie = (COOKIE **) ap_push_array(cookies);

        /* Allocating the COOKIE structure */
        *cookie = (COOKIE *) ap_palloc(r->pool, sizeof(COOKIE));

        /* Retrieving the cookie name before the '=' sign    */
        /* After that, pair will contain the key/value pairs */
        name = ap_getword(r->pool, &pair, '=');
        /* OR 6820 : remove trailing spaces */
        if (name[0] != '\0')
        {
            char *end = (char *) name + strlen(name) - 1;
            while (end >= name && end[0] == ' ')
            {
                end[0] = '\0';
                end--;
            }
        }
        /* Skipping spaces between equal and cookie value */
        while (ap_isspace(*pair)) ++pair;
/*        ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "cookie name %s", name);*/
        /*ap_unescape_url((CSICHAR *)pair);*/
        (*cookie)->name = ap_pstrdup(r->pool, name);
        (*cookie)->size = strlen(pair);
        /* Parsing the key/value pairs from the cookie */
        (*cookie)->pair = ap_make_table(r->pool, 2);
        while (   NULL != pair
	       && *pair
               && (key = ap_getword_nulls(r->pool, &pair, '&'))
              )
        {
            ap_unescape_url((char *) key);
/*            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "cookie key : %s", key);*/
            value = ap_getword_nulls(r->pool, &pair, '&');
            ap_unescape_url((char *) value);
/*            ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r, "cookie unescaped value : %s", value);*/
            ap_table_add((*cookie)->pair, key, value);
        }
    }

    return cookies;
}


table *CookieFind(request_rec *r
                 ,const char *name
                 ,array_header *cookies
                 )
{
    CSILONG i;
    CSIBOOLEAN found;
    COOKIE **list;

    if (NULL == cookies)
        return NULL;

    list = (COOKIE **) cookies->elts;
 
    for(i = 0, found = CSIFALSE
       ;i < cookies->nelts && !found
       ;i++
       )
    {
        COOKIE *cookie;

        cookie = list[i];
        found = strcmp(cookie->name, name) ? CSIFALSE : CSITRUE;
        if (found)
            return cookie->pair;
    } 

    return NULL;
}
