/****************************************************************************

         #####   #####  #     #       #  #####   #####    ###
        #     # #     # ##    #      #  #     # #     #    #
        #       #       # #   #     #   #       #          #
        #       #       #  #  #    #    #        #####     #
        #       #       #   # #   #     #             #    #
        #     # #     # #    ##  #      #     # #     #    #
         #####   #####  #     # #        #####   #####    ###

****************************************************************************/
/*
* @PROJECT  CCN/CSI
* @SYSTEM   APACHE
* @PACKAGE  RAPMGMT 
*
* @FILE     rapmgmt_exe.c
*
* @ABSTRACT
* Apache RAP ManaGeMenT (RAPMGMT) : C program launched by the handler
*
* This file contains:
* the C program that will be launched by the Apache handler
*
* @INCLUDE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <stdarg.h>
#include <errno.h>
#include <signal.h>
#include <values.h>

#include "gcn_uptyp.h"
#include "gcn_upcon.h"
#include "osx_upcon.h"
#include "osx_upext.h"
#include "cdx_upcon.h"
#include "cdx_uptyp.h"
#include "cdx_upmac.h"
#include "cdx_upext.h"

#include "com_upext.h"
#include "rapmgmt_prtyp.h"
#include "rapmgmt_prcon.h"
#include "rapmgmt_prmes.h"
#include "rapmgmt_upext.h"

extern CSICHAR *szAction;
extern CSICHAR *szSiteString;
extern CSICHAR *szUserSite;
extern CSICHAR *szGtwString;
extern CSICHAR *szRapString;
extern CSICHAR *szProfileString;
extern CSICHAR *szAppDomString;

extern CSICHAR48 szWebDomain;
extern CSIBOOLEAN OSInit;
extern CSIBOOLEAN CDXBind;

RAPLIST *psRapList = NULL;
CDXSITEATTINFO sSiteAtt;
CDXGTWATTINFO sGatewayAtt;


CSIVOID FreeRapList(CSIVOID)
{
    RAPLIST *psRapItem = NULL;
    CSILONG lReturnCode;
    CSILONG lReasonCode;

    while (NULL != psRapList)
    {
        psRapItem = psRapList;
        psRapList = psRapItem->psNext;
        OS_mem_free((CSIVOID **) &psRapItem, &lReturnCode, &lReasonCode);
    }
}

static CSIVOID HandleError(CSILONG lErrorId, ...)
{
    va_list arguments;
    CSICHAR1024 szMsg;
    CSILONG lReturnCode;
    CSILONG lReasonCode;

    va_start(arguments, lErrorId);
    vsprintf(szMsg
            ,RAPMGMT_ERROR_MSGS[lErrorId - 1]
            ,arguments
            );
    va_end(arguments);

    PrintAdminServiceBegin(CSIFAILED);
    PrintErrorMessage(lErrorId, szMsg);
    PrintAdminServiceEnd();


    FreeRapList();

    if (CSITRUE == CDXBind)
    {
        CDX_unbind(NULL, &lReturnCode, &lReasonCode);
        CDXBind = CSIFALSE;
    }

    if (CSITRUE == OSInit)
    {
        OS_term(&lReturnCode, &lReasonCode);
        OSInit = CSIFALSE;
    }

    exit(1);
}


static void 
SignalHandling(int signo)
{
    if (signo == SIGPIPE)
    {
        /* Error while trying to read from pipe */
        HandleError(RAPMGMT_ERROR_PIPE_SIGNAL);
    }
    else if (signo == SIGCHLD)
    {
        pid_t pid;
        int status;

        if ( (pid = waitpid((pid_t) -1, &status, WUNTRACED)) < 0)
        {   
            HandleError(RAPMGMT_ERROR_WAITPID, errno);
        }   
        if (WIFEXITED(status))
        {
            int childstatus = WEXITSTATUS(status);
            if (childstatus != 0)
            {
                HandleError(RAPMGMT_ERROR_CHILD, (CSILONG) pid, childstatus);
            }
            else if (WIFSIGNALED(status))
            {
                HandleError(RAPMGMT_ERROR_CHILD_SIGNALED, (CSILONG) pid, WTERMSIG(status));
            }
            else if (WIFSTOPPED(status))
            {
                HandleError(RAPMGMT_ERROR_CHILD_STOPPED, (CSILONG) pid, WSTOPSIG(status));
            }
            /* No error */
            return;
        }
    }
}


CSIVOID RapListPrint(CSIVOID)
{
    RAPLIST *psRapItem;

    PrintAdminServiceBegin(CSISUCCESS);
    PrintDomainBegin(szWebDomain);
    PrintSiteBegin(2, sSiteAtt);
    PrintGatewayBegin(3, sGatewayAtt);
    
    for (psRapItem = psRapList
        ;NULL != psRapItem
        ;psRapItem = psRapItem->psNext
        )
    {
        if (   (   psRapItem->CanBeListed  
                && strcmp(szAction, RAPMGMT_CHECK) == 0
               )
            || psRapItem->CanBeManaged
           )
        {
            CSICHAR16 szRapState;

            if (RAPSTATE_RUNNING == psRapItem->Running)
                strcpy(szRapState, RAPSTATE_RUNNING_STR);
            else if (RAPSTATE_STOPPED == psRapItem->Running)
                strcpy(szRapState, RAPSTATE_STOPPED_STR);
            else
                strcpy(szRapState, RAPSTATE_UNKNOWN_STR);
 
            PrintRapMgmtRap(psRapItem->szRapName
                           ,psRapItem->szAppdName
                           ,psRapItem->szMode
			   ,psRapItem->CanBeManaged
                           ,szRapState
                           );
        }
    }

    PrintGatewayEnd(3);
    PrintSiteEnd(2);
    PrintDomainEnd();
    PrintAdminServiceEnd();
}


static CSIVOID AddRapItem(RAPLIST *psRapItem)
{
    RAPLIST *psPrev;

    /* First place in the list */
    if (   psRapList == NULL
        || strcmp(psRapItem->szRapName, psRapList->szRapName) < 0
       )
    {   
        psRapItem->psNext = psRapList;
        psRapList = psRapItem;
        return;
    }
    psRapItem->psNext = psRapList->psNext;
    psRapList->psNext = psRapItem;
    psPrev = psRapList;
    while (   psRapItem->psNext
           && strcmp(psRapItem->psNext->szRapName, psRapItem->szRapName) < 0
          )
    {
        psPrev->psNext = psRapItem->psNext;
        psPrev = psRapItem->psNext;
        psRapItem->psNext = psPrev->psNext;
        psPrev->psNext = psRapItem;
    }
}


CSIVOID BuildRapList(CSIVOID)
{
    CSILONG lReturnCode;
    CSILONG lReasonCode;
    CSILONG lNbProfile;
    CSICHAR **szProfile;
    CSICHAR *szString;
    CSICHAR128 szFilter = "";
    CDXRESULT sCDXAppList = {CDXRESULT_DEFAULT};
    CSICHAR *szAppName;
    CSILONG i;

    if (CSISUCCESS != CDX_getAttributesOfSite(NULL
                                             ,szSiteString
                                             ,&sSiteAtt   
                                             ,&lReturnCode
                                             ,&lReasonCode
                                             )            
       )
    {   
        HandleError(RAPMGMT_ERROR_CDX
                   ,"CDX_getAttributesOfSite"
                   ,lReturnCode
                   ,lReasonCode
                   );          
    }
     
    if (CSISUCCESS != CDX_getAttributesOfGateway(NULL
                                                ,szSiteString
                                                ,szGtwString 
                                                ,&sGatewayAtt
                                                ,&lReturnCode
                                                ,&lReasonCode
                                                )            
       )
    {   
        HandleError(RAPMGMT_ERROR_CDX
                   ,"CDX_getAttributesOfGateway"
                   ,lReturnCode
                   ,lReasonCode
                   );          
    }
     
    /* For a STOP/START operation, the filter is set to the   */
    /* name of the RAP. It is combined in the function itself */
    /* with ccnApplicationType set to External.               */
    if (!strcmp(szAction, RAPMGMT_START) || !strcmp(szAction, RAPMGMT_STOP))
        snprintf(szFilter
                ,sizeof(szFilter)
                ,"(rap=%s)"
                ,szRapString
                );          
    if (CSISUCCESS != CDX_listApplicationsOfGateway(NULL
                                                   ,szSiteString
                                                   ,szGtwString 
                                                   ,szFilter   
                                                   ,&sCDXAppList
                                                   ,&lReturnCode
                                                   ,&lReasonCode
                                                   )            
       )
    {   
        HandleError(RAPMGMT_ERROR_CDX
                   ,"CDX_listApplicationsOfGateway"
                   ,lReturnCode
                   ,lReasonCode
                   );          
    }
     
    for (szAppName = CDX_list_first(NULL, &sCDXAppList)
        ;szAppName
        ;szAppName = CDX_list_next(NULL, &sCDXAppList)
        )
    {
        CDXAPPATTINFO sAppAtt;
        RAPLIST *psRapItem;
        CSICHAR *szGtwMode;
        CSIBOOLEAN found;

        if (CSISUCCESS != CDX_getAttributesOfApplication(NULL
                                                        ,szSiteString
                                                        ,szGtwString 
                                                        ,szAppName  
                                                        ,&sAppAtt 
                                                        ,&lReturnCode
                                                        ,&lReasonCode
                                                        )            
           )
        {   
            HandleError(RAPMGMT_ERROR_CDX
                       ,"CDX_getAttributesOfApplication"
                       ,lReturnCode
                       ,lReasonCode
                       );          

        }
         
        /* No activation mode defined */
        if ((sAppAtt.lFields & CDX_APPATT_ACTMODE) != CDX_APPATT_ACTMODE)
            continue;
        /* Activation mode set to initiator */
        if (   (sAppAtt.lFields & CDX_APPATT_ACTMODE) == CDX_APPATT_ACTMODE
            && strcasecmp(sAppAtt.szActivationMode, CDX_C_INITIATOR_STR) == 0
           )
            continue;
        /* No RAP defined */
        if ((sAppAtt.lFields & CDX_APPATT_RAPNAME) != CDX_APPATT_RAPNAME)
            continue;
        /* No mode defined */
        if ((sAppAtt.lFields & CDX_APPATT_MODE) != CDX_APPATT_MODE)
            continue;
        /* The RAP must be in one of the mode of the gateway */
        for (found = CSIFALSE, szGtwMode = CDX_list_first(NULL, &sGatewayAtt.sGtwModeList)
            ;szGtwMode != NULL && found == CSIFALSE
            ;szGtwMode = CDX_list_next(NULL, &sGatewayAtt.sGtwModeList)
            )
        {    
            found = (strcasecmp(szGtwMode, CDX_GTWMODE_CSICHAR(sAppAtt.lRunMode)) == 0)
                        ? CSITRUE
                        : CSIFALSE;
        }
        if (found == CSIFALSE)
            continue;
        /* The RAP can be added to the list */
        if (CSISUCCESS != OS_mem_alloc(sizeof(RAPLIST)
                                      ,(CSIVOID **) &psRapItem
                                      ,&lReturnCode           
                                      ,&lReasonCode
                                      )            
           )
        {   
            HandleError(RAPMGMT_ERROR_OS
                       ,"OS_mem_alloc"  
                       ,lReturnCode   
                       ,lReasonCode
                       );          
        }
         
        memset(psRapItem, '\0', sizeof(RAPLIST));
        strcpy(psRapItem->szRapName, sAppAtt.szRapName);
        strcpy(psRapItem->szAppdName, strrchr(psRapItem->szRapName, '.') + 1);
        strcpy(psRapItem->szAppdName, strrchr(psRapItem->szRapName, '.') + 1);
        strcpy(psRapItem->szMode, CDX_GTWMODE_CSICHAR(sAppAtt.lRunMode));
        psRapItem->CanBeListed = CSIFALSE;
        psRapItem->CanBeManaged = CSIFALSE;
        psRapItem->Running = 0;
        AddRapItem(psRapItem);
    }
     
    CDX_free(NULL, &sCDXAppList);

    lNbProfile = 1;
    szString = szProfileString;
    while ( (szString = strchr(szString, ' ')) != NULL)
    {
        szString++;
        lNbProfile++;
    }
     
    lReturnCode = CSISUCCESS;
    lReasonCode = OS_ERR_NONE;
    if (CSISUCCESS != OS_mem_alloc(lNbProfile * sizeof(CSICHAR *)
                                  ,(CSIVOID **) &szProfile       
                                  ,&lReturnCode           
                                  ,&lReasonCode
                                  )            
       )
    {   
        HandleError(RAPMGMT_ERROR_OS
                   ,"OS_mem_alloc"  
                   ,lReturnCode   
                   ,lReasonCode
                   );          
    }
     
    szString = szProfile[0] = szProfileString;
    i = 0;
    while ( (szString = strchr(szString, ' ')) != NULL)
    {
        szString[0] = '\0';
        szString++;
        i++;
        szProfile[i] = szString;
    }
     
    for (i = 0; i < lNbProfile ; i++)
    {
        CDXRESULT adminToolList;
        CDXRESULT appDomainList;
        CDXRESULT usageModeList;
        CSICHAR48 szAppdName;
        CSILONG lRights;
        CSILONG lRole;
        RAPLIST *psRapItem;
        CSICHAR *szTemp;
        CSICHAR *pcTool;
        CSIBOOLEAN adminTool = CSIFALSE;

        if (CSISUCCESS != CDX_getAttributesOfProfile(NULL
                                                    ,szProfile[i]
                                                    ,&adminToolList
                                                    ,&appDomainList
                                                    ,&lRights      
                                                    ,&lRole  
                                                    ,&usageModeList
                                                    ,&lReturnCode  
                                                    ,&lReasonCode
                                                    )            
           )
        {   
            HandleError(RAPMGMT_ERROR_CDX
                       ,"CDX_getAttributesOfProfile"
                       ,lReturnCode
                       ,lReasonCode
                       );          
        }
         
        *szAppdName = '\0';
        if (szTemp = CDX_list_first(NULL, &appDomainList))
           strcpy(szAppdName, szTemp);

        CDX_free(NULL, &appDomainList);

        pcTool  = CDX_list_first(NULL, &adminToolList);
        while (pcTool != NULL)
        {
            if (!strcmp(pcTool, CDX_C_RAPMGMT_STR))
            {
                adminTool = CSITRUE;
            }
            pcTool  = CDX_list_next(NULL, &adminToolList);
        }
         
        CDX_free(NULL, &adminToolList);

        szTemp = CDX_list_first(NULL, &usageModeList);

        if (!*szAppdName || !szTemp || !*szTemp || adminTool == CSIFALSE)
        {
            CDX_free(NULL, &usageModeList);
            continue;
        }
         
        if (CDX_C_LCMASTER != lRole && CDX_C_CCMASTER != lRole)
            continue;

        if (CDX_C_LIST != lRights && CDX_C_MGMT != lRights)
            continue;

        for (psRapItem = psRapList
            ;NULL != psRapItem
            ;psRapItem = psRapItem->psNext
            )
        {    
            CSICHAR    *szUsageMode;
            CSIBOOLEAN lUsageMode = CSIFALSE;

            for (szUsageMode = CDX_list_first(NULL, &usageModeList)
                ;szUsageMode != NULL && CSIFALSE == lUsageMode
                ;szUsageMode = CDX_list_next(NULL, &usageModeList)
                )
            {    
                lUsageMode = (strcasecmp(szUsageMode, psRapItem->szMode) == 0)
                                 ? CSITRUE
                                 : CSIFALSE;
            }
             
            if (!psRapItem->CanBeListed)
                psRapItem->CanBeListed = (   (   CDX_C_LIST == lRights
                                              || CDX_C_MGMT == lRights
                                             )                        
                                          && lUsageMode
                                          && (   CDX_C_CCMASTER == lRole
                                              || (   CDX_C_LCMASTER == lRole
                                                  && (!strcmp(szUserSite, szSiteString))
                                                 )                                      
                                             )    
                                          && (!strcmp(szAppdName, psRapItem->szAppdName))
                                          && (   strcmp(szAction, RAPMGMT_CHECK)
                                              || !strcmp(szAppDomString, psRapItem->szAppdName)
                                              || !strcmp(szAppDomString, RAPMGMT_APPD_ALL)
                                             )                                            
                                         );   

            if (!psRapItem->CanBeManaged)
                psRapItem->CanBeManaged = (   CDX_C_MGMT == lRights
                                           && lUsageMode           
                                           && (   CDX_C_CCMASTER == lRole
                                               || (   CDX_C_LCMASTER == lRole
                                                   && (!strcmp(szUserSite, szSiteString))
                                                  )                                      
                                              )    
                                           && (!strcmp(szAppdName, psRapItem->szAppdName))
                                          && (   strcmp(szAction, RAPMGMT_CHECK)
                                              || !strcmp(szAppDomString, psRapItem->szAppdName)
                                              || !strcmp(szAppDomString, RAPMGMT_APPD_ALL)
                                             )                                            
                                          );  
        }
    }
     
    OS_mem_free((CSIVOID **) &szProfile, &lReturnCode, &lReasonCode);
}


static RAPLIST *FindRap(CSIVOID)
{
    RAPLIST *psRapItem = NULL;

    psRapItem = psRapList;
    while (   psRapItem != NULL
           && strcmp(szRapString,  psRapItem->szRapName) != 0
          )
    {
        psRapItem = psRapItem->psNext;
    }

    return (psRapItem);
}


CSIVOID StartStopRap(CSIVOID)
{
    int pfd1[2], pfd2[2];
    pid_t pid;
    RAPLIST *psRapItem;


    psRapItem = FindRap();
    if (NULL == psRapItem)
    {
        HandleError(RAPMGMT_ERROR_RAP_NOT_FOUND
                   ,szRapString
                   ,szGtwString 
                   ,szSiteString 
                   );           
    }

    if (!psRapItem->CanBeManaged)
    {
        HandleError(RAPMGMT_ERROR_RAP_MANAGED
                   ,szRapString
                   ,szGtwString
                   ,szSiteString 
                   );          
    }

    if (   signal(SIGPIPE, SignalHandling) == SIG_ERR
        || signal(SIGCHLD, SignalHandling) == SIG_ERR
       )
    {
        HandleError(RAPMGMT_ERROR_SIGNAL, errno);          
    }

    if (pipe(pfd1) < 0 || pipe(pfd2))
    {
        HandleError(RAPMGMT_ERROR_PIPE, errno);          
    }

    if ( (pid = fork()) < 0)
    {
        HandleError(RAPMGMT_ERROR_FORK, errno);          
    }

    if (pid == 0)
    {
        /* son */
        CSICHAR256 szCmd;
        CSICHAR szLibPath[MAXLINE];

        close(pfd1[1]);
        close(pfd2[0]);

        if (STDOUT_FILENO != pfd2[1])
        {
            if (STDOUT_FILENO != dup2(pfd2[1], STDOUT_FILENO))
            {
                HandleError(RAPMGMT_ERROR_DUP2, errno);
            }
        }

        if (STDERR_FILENO != pfd2[1])
        {
            if (STDERR_FILENO != dup2(pfd2[1], STDERR_FILENO))
            {
                HandleError(RAPMGMT_ERROR_DUP2, errno);
            }
            if (STDOUT_FILENO != pfd2[1])
                close(pfd2[1]);
        }

        if (STDIN_FILENO != pfd1[0])
        {
            if (STDIN_FILENO != dup2(pfd1[0], STDIN_FILENO))
            {
                HandleError(RAPMGMT_ERROR_DUP2, errno);
            }
            close(pfd1[0]);
        }

        snprintf(szCmd
                ,sizeof(szCmd)
                ,"%s/bin/%s"
                ,getenv("TUXDIR") == NULL
                     ? ""
                     : getenv("TUXDIR")
                ,TUXEDO_TMADMIN
                );
        snprintf(szLibPath, sizeof(szLibPath), 
#ifdef CCN_LINUX
                                               "LD_LIBRARY_PATH=%s", getenv("LD_LIBRARY_PATH"));
#else /* CCN_LINUX */
#ifdef CCN_AIX
                                               "LIBPATH=%s", getenv("LIBPATH"));
#else /* CCN_AIX */
#error Neither CCN_LINUX or CCN_AIX defined in compilation flags
#endif /* CCN_AIX */
#endif /* CCN_LINUX */


        if (0 > execlp("sudo"
                      ,"sudo", "-u", "tuxedo"
                      ,"env", szLibPath
                      ,szCmd
                      ,(char *) 0
                      )
           )
        {
            HandleError(RAPMGMT_ERROR_EXEC, errno);
        }
        exit(1);
    }
    else
    {
        CSICHAR1024 szTuxedo;
        CSILONG lLength = 0;

        /* father */
        close(pfd1[0]);
        close(pfd2[1]);

        memset(szTuxedo, '\0', sizeof(szTuxedo));

        if (!strcmp(szAction, RAPMGMT_STOP))
        {
            lLength = sprintf(szTuxedo
                             ,TUXEDO_STOP
                             ,szRapString
                             );
        }
        else if (!strcmp(szAction, RAPMGMT_START))
        {
            lLength = sprintf(szTuxedo
                             ,TUXEDO_START
                             ,szRapString
                             );
            
        }

        if (write(pfd1[1], szTuxedo, lLength) != lLength)
        {
            HandleError(RAPMGMT_ERROR_WRITE, errno);
        }

        /* waitpid() is now replaced by a sleep() that will */
        /* be interrupted by a SIGPIPE or SIGCHLD */
        sleep(MAXINT);
    }
}


CSIVOID CheckRap (CSIVOID)
{
    RAPLIST *psRapItem;
    int pfd1[2], pfd2[2];
    pid_t pid;

    if (   signal(SIGPIPE, SignalHandling) == SIG_ERR
        || signal(SIGCHLD, SignalHandling) == SIG_ERR
       )
    {
        HandleError(RAPMGMT_ERROR_SIGNAL, errno);          
    }

    if (pipe(pfd1) < 0 || pipe(pfd2))
    {
        HandleError(RAPMGMT_ERROR_PIPE, errno);
    }

    if ( (pid = fork()) < 0)
    {
        HandleError(RAPMGMT_ERROR_FORK, errno);
    }

    if (pid == 0)
    {
        /* child */
        int devnull;

        close(pfd1[1]);
        close(pfd2[0]);

        if (STDOUT_FILENO != pfd2[1])
        {
            if (STDOUT_FILENO != dup2(pfd2[1], STDOUT_FILENO))
            {
                HandleError(RAPMGMT_ERROR_DUP2, errno);
            }
            close(pfd2[1]);
        }    
        
        devnull = open("/dev/null", O_RDWR); 
        if (STDERR_FILENO != devnull)
        {
            if (STDERR_FILENO != dup2(devnull, STDERR_FILENO))
            {
                HandleError(RAPMGMT_ERROR_DUP2, errno);
            }
            close(devnull);
        }

        if (STDIN_FILENO != pfd1[0])
        {
            if (STDIN_FILENO != dup2(pfd1[0], STDIN_FILENO))
            {
                HandleError(RAPMGMT_ERROR_DUP2, errno);
            }
            close(pfd1[0]);
        }

        if (0 > execlp(TUXEDO_RAPCHECK
                      ,TUXEDO_RAPCHECK
                      ,(char *) 0
                      )
           )
        {
            HandleError(RAPMGMT_ERROR_EXEC, errno);
        }
    }

    /* father */
    close(pfd1[0]);
    close(pfd2[1]);

    for (psRapItem = psRapList
        ;NULL != psRapItem
        ;psRapItem = psRapItem->psNext
        )
    {

        if (   psRapItem->CanBeListed == CSITRUE
            || psRapItem->CanBeManaged == CSITRUE
           )
        {
            /* The status of the rap must be checked */
            int n, index = 0;
            char c = '\0';
            CSICHAR48 szLine;

            /* sending the name of the rap to the Tuxedo program */
            n = strlen(psRapItem->szRapName);
            if (   n != write(pfd1[1], psRapItem->szRapName, n)
                || 1 != write(pfd1[1], "\n", 1)
               )
            {   
                HandleError(RAPMGMT_ERROR_WRITE, errno);
            }   

            /* Retrieving the status of the rap */
            memset(szLine, '\0', sizeof(szLine));
            while (c != '\n')
            {
                if (read(pfd2[0], &c, 1) < 0)
                {
                    HandleError(RAPMGMT_ERROR_READ, errno);
                }
                szLine[index++] = c;
            }
            szLine[index - 1] = '\0';

            psRapItem->Running = (!strcmp(szLine, TUXEDO_STATE_ACTIVE)) 
                               ? RAPSTATE_RUNNING
                               : RAPSTATE_STOPPED;
        }
    }
    close(pfd1[1]);
    close(pfd2[0]);
}
